__version__ = "2.3.0"
# If you're making a patch or a minor bug fix, increment the patch version,
# e.g., from 0.1.0 to 0.1.1.
# If you're adding functionality in a backwards-compatible manner, increment
# the minor version, e.g., from 0.1.0 to 0.2.0.
# If you're making incompatible API changes, increment the major version, e.g.,
# from 0.1.0 to 1.0.0.
